﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Imaging;         //【Bitmap】
using System.Runtime.InteropServices; //【Debug】
using System.Diagnostics;             //【Debug】
using Microsoft.Kinect;               //【Kinect】
using System.Runtime.InteropServices; //【Marshal】
using System.IO;                      //【FileStream】
using NTL.KinectLib;                  //【ColorBar】

namespace K4W_v2_NTLKinectFusion3
{
    public partial class Form1 : Form
    {
        private KinectSensor KS = null;                    //【KinectSensor】１台のみ使用可
        private MultiSourceFrameReader multiReader = null; //【フレームデータを読み込むためのReader】
        private const int nFrame = 100;                    //1秒分(30フレーム)位が良好
        private float Kd;                                  //平均を求めるための乗算係数;
        private const int Dx = 512;
        private const int Dy = 424;
        private const int dByte = 4;
        private const int dPixels = Dx * Dy;
        private byte[] dXRGB1 = new byte[dPixels * dByte];         //【表示用(OriginalDepth)】
        private byte[] dXRGB2 = new byte[dPixels * dByte];         //【表示用(AverageDepth)】
        private byte[] dXRGB3 = new byte[dPixels * dByte];         //【表示用(AverageDepth)】
        private Bitmap dBmp1 = new Bitmap(Dx, Dy, PixelFormat.Format32bppRgb); //【原Depth】512×424画素
        private Bitmap dBmp2 = new Bitmap(Dx, Dy, PixelFormat.Format32bppRgb); //【平均Depth】512×424画素
        private Bitmap dBmp3 = new Bitmap(Dx, Dy, PixelFormat.Format32bppRgb); //【平均Depth】512×424画素
        private Bitmap dBmp1s = new Bitmap(256, 212, PixelFormat.Format32bppRgb);
        private Bitmap dBmp2s = new Bitmap(256, 212, PixelFormat.Format32bppRgb);
        private Rectangle dRect = new Rectangle(0, 0, Dx, Dy);
        private BitmapData dBmpData = null;
        private ColorBar depthBar = new ColorBar();
        private int ptr = 0;                                       //【DepthFramFIFOのFrameポインタ】
        private int nAve = 0;                                      //【有効加算回数】
        private ushort[] DataIn = new ushort[dPixels];             //【一時バッファ】
        private ushort[] DepthFIFO = new ushort[dPixels * nFrame]; //【DepthFrameFIFO】
        private long[] Sum = new long[dPixels];                    //【移動平均演算用Sum】
      //  private int Count = 0;
        private Boolean flag = false;


        //**********************************************************
        //【メイン】
        //**********************************************************
        public Form1()
        {
            InitializeComponent();

            this.Load += Form1_Load;               //【Kinectの初期化を行うイベント】
            this.FormClosing += Form1_FormClosing; //【Kinectのリソースを解放するイベント】

            timer1.Interval = 33;
            timer1.Tick += timer1_Tick;
            timer1.Enabled = true;
          //  timer1.Enabled = false;
        }

        //**********************************************************
        //【初期化処理】
        //**********************************************************
        void Form1_Load(object sender, EventArgs e)
        {
            KS = KinectSensor.GetDefault(); //【Kinect取得】
            if (KS == null) { return; }     //●取得失敗なら何も行わない
            multiReader = KS.OpenMultiSourceFrameReader(FrameSourceTypes.Depth);        //【DepthFrameのみ】
            multiReader.MultiSourceFrameArrived += multiReader_MultiSourceFrameArrived; //【フレーム読み込みイベント】
            KS.Open();                      //【Kinect起動】

            depthBar.CreateColorPalletForDistance();                                       //【距離カラーマップを生成】
            Bitmap bmp = depthBar.CreateColorScale(65536, 10, false);                      //【距離カラーマップゲージ】16bit
            pictureBox3.Image = bmp.Clone(new Rectangle(0, 0, 4500, 10), bmp.PixelFormat); //【距離カラーマップゲージを切り取る】

            FIFOInitialize();               //【DepthFrameFIFOの初期化処理】
            NTL_PixelShaderInitialize();
            label1.Text = ""; label12.Text = ""; label13.Text = ""; label14.Text = "";
        }

        //**********************************************************
        //【終了処理】
        //**********************************************************
        void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (multiReader != null) { multiReader.Dispose(); multiReader = null; }
            if (KS != null) { KS.Close(); KS = null; }
            //【表示系】
             dBmp1.Dispose(); dBmp2.Dispose(); dXRGB1 = dXRGB2 = null;
            //【DepthFrameFIFO系】
            DataIn = null; DepthFIFO = null; Sum = null; Vertex = null; Indices = null;
        }

        //**********************************************************
        //【フレーム画像取得処理】
        //**********************************************************
        void multiReader_MultiSourceFrameArrived(object sender, MultiSourceFrameArrivedEventArgs e)
        {
            MultiSourceFrame multiFrame = e.FrameReference.AcquireFrame();
            
            //【カラーフレーム 512×424画素 15FPS】
            using (DepthFrame depthFrame = multiFrame.DepthFrameReference.AcquireFrame())
            {
                if (depthFrame != null)
                {
                    depthFrame.CopyFrameDataToArray(DataIn); //【デプスデータ取得】Depthデータ⇒DataIn
                    ConvertDepth();                          //【FIFOで平均処理とカラー変換】デプス値,平均デプス値⇒Color

                    dBmpData = dBmp1.LockBits(dRect, ImageLockMode.WriteOnly, PixelFormat.Format32bppRgb); //【Lockする】
                    Marshal.Copy(dXRGB1, 0, dBmpData.Scan0, dXRGB1.Length); //【データ転送】メモリ⇒Bitmap
                    dBmp1.UnlockBits(dBmpData);                                                            //【Lock解除】
                    Graphics g1 = Graphics.FromImage(dBmp1s);
                    g1.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic;
                    g1.DrawImage(dBmp1, 0, 0, dBmp1s.Width, dBmp1s.Height);
                    dBmp1s.RotateFlip(RotateFlipType.Rotate180FlipY); //【左右反転】
                    pictureBox1.Image = dBmp1s;                       //【pictureBox1に画像表示】
                    g1.Dispose();
                    
                    dBmpData = dBmp2.LockBits(dRect, ImageLockMode.WriteOnly, PixelFormat.Format32bppRgb); //【Lockする】
                    Marshal.Copy(dXRGB2, 0, dBmpData.Scan0, dXRGB2.Length); //【データ転送】メモリ⇒Bitmap
                    dBmp2.UnlockBits(dBmpData);                                                            //【Lock解除】
                    Graphics g2 = Graphics.FromImage(dBmp2s);
                    g2.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic;
                    g2.DrawImage(dBmp2, 0, 0, dBmp2s.Width, dBmp2s.Height);
                    dBmp2s.RotateFlip(RotateFlipType.Rotate180FlipY); //【左右反転】
                    pictureBox2.Image = dBmp2s;                       //【pictureBox1に画像表示】
                    g2.Dispose();
                }
            }
        }


        

        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        //【DepthFrameFIFOの初期化】
        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        private void FIFOInitialize()
        {
            ptr = 0; nAve = 0;
            for (int i = 0; i < DepthFIFO.Length; i++) { DepthFIFO[i] = 0; }
            for (int i = 0; i < Sum.Length; i++) { Sum[i] = 0; }
        }

        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        //【DepthFrameFIFOのフレーム単位の処理】
        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        private void ConvertDepth()
        {
            int j = 0;
            ushort depth;
            int L = dPixels * ptr;
            nAve++;
            if (nAve > nFrame) { nAve = nFrame; } else { Kd = 1.0f / (float)nAve; }
            float z; const float zb = 512.0f;
            int x = 0; int y = 0; const int xc = 256; const int yc = 212;
            for (int i = 0; i < dPixels; i++)
            {
                //【FIFO & MovingAverage】
                depth = DataIn[i];
                Sum[i] += (long)depth - (long)DepthFIFO[L + i];
                DepthFIFO[L + i] = depth;

                //【Display】
                dXRGB1[j] = depthBar.B[depth];
                dXRGB1[j + 1] = depthBar.G[depth];
                dXRGB1[j + 2] = depthBar.R[depth];
                dXRGB1[j + 3] = 0;
                depth = (ushort)(Sum[i] * Kd);
                dXRGB2[j] = depthBar.B[depth];
                dXRGB2[j + 1] = depthBar.G[depth];
                dXRGB2[j + 2] = depthBar.R[depth];
                dXRGB2[j + 3] = 0;
                j += 4;

                //【Create Vertex】≪頂点データ≫
                z = (float)Sum[i] * Kd *0.001f; // Unit [mm]⇒[m]
                if ((z < 0.5f) || (z > 4.5f)) { z = 5.0f; } // for NormalVector
                Vertex[i].X = -((float)(x - xc) * z / zb); 
                Vertex[i].Y = ((float)(y - yc) * z / zb); 
                Vertex[i].Z = z;
                x++; if (x == Dx) { x = 0; y++; }
            }
            ptr++; if (ptr == nFrame) { ptr = 0; } //【ポインタ更新】
        }


        //**********************************************************
        //【タイマーイベント】
        //**********************************************************
        void timer1_Tick(object sender, EventArgs e)
        {
            NTL_PixelShader();

            if (flag == false) return;
            CreateSTLFile(); //
            label1.Text = "Numbers of Polygons: " + nPolygon.ToString() + "\nNumbers of Indecies: " + nIndex.ToString();
            label12.Text = "Xmin=" + Xmin.ToString() + "\nXmax=" + Xmax.ToString();
            label13.Text = "Ymin=" + Ymin.ToString() + "\nYmax=" + Ymax.ToString();
            label14.Text = "Zmin=" + Zmin.ToString() + "\nZmax=" + Zmax.ToString();
            flag = false;

        }



        //**********************************************************
        //【】
        //**********************************************************
        private void button1_Click(object sender, EventArgs e)
        {
            flag = true;
        }


        private void button2_Click(object sender, EventArgs e)
        {
            flag = true;
        }


        //**********************************************************
        //【NTL Pixel Shader】
        //
        //
        //
        //
        //**********************************************************
        public struct NTLVector3
        {
            public float X;
            public float Y;
            public float Z;
        }

        public struct NTLRGB
        {
            public float R;
            public float G;
            public float B;
        }


        public struct D3D_Pixel
        {
            public int Width;
            public int Height;
            public float[] Depth;
            public NTLVector3[] Normal;
            public NTLVector3 Light;
            public NTLVector3 Camera;
            public NTLRGB FaceColor;
            public NTLRGB DiffuseColor;
            public float Power;
            public NTLRGB SpecularColor;
        }

        //【STLファイル作成関連】
        private NTLVector3[] Vertex = new NTLVector3[dPixels];        //【頂点座標】
        private int[] Indices = new int[3 * 2 * (Dx - 1) * (Dy - 1)]; //【インデックス】
        private int nIndex;   //【ポリゴン数×3】
        private int nPolygon; //【ポリゴン数】5cm以上長い辺をもったポリゴンは削除

        private float Xmin, Xmax; //【デバッグ用】特に必要ではない
        private float Ymin, Ymax; //【デバッグ用】特に必要ではない
        private float Zmin, Zmax; //【デバッグ用】特に必要ではない

        //【NTL Pixel Shader】
        private D3D_Pixel Pix;

        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        //【NTL_Pixel Shaderの初期化】
        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        void NTL_PixelShaderInitialize()
        {
            Pix.Width = 512 - 2;
            Pix.Height = 424 - 2;
            Pix.Normal = new NTLVector3[Pix.Width * Pix.Height];
            //【視線ベクトル】正規化
            Pix.Camera.X = 0.0f; Pix.Camera.Y = 0.0f; Pix.Camera.Z = 1.0f;
            Pix.Camera = VectorNormailze(Pix.Camera);
            //【光源ベクトル】正規化　●光源の方向を示す事に注意！
            Pix.Light.X = 1.0f; Pix.Light.Y = -1.0f; Pix.Light.Z = -1.0f;
            Pix.Light = VectorNormailze(Pix.Light);
            //【材質】環境光, 拡散光, 鏡面反射光の各成分
            Pix.FaceColor.R = 0.0f; Pix.FaceColor.G = 0.0f; Pix.FaceColor.B = 1.0f;
            Pix.DiffuseColor.R = 0.4f; Pix.DiffuseColor.G = 0.8f; Pix.DiffuseColor.B = 1.0f;
            Pix.Power = 5.0f;
            Pix.SpecularColor.R = 0.5f; Pix.SpecularColor.G = 0.5f; Pix.SpecularColor.B = 0.5f;
            //画像用配列をゼロクリア
            for (int i = 0; i < dXRGB3.Length; i++ ) { dXRGB3[i] = 0; }
        }

        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        //【NTL_Pixel Shaderの初期化】
        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        void NTL_PixelShader()
        {
            NTLVector3 L = Pix.Light;  //【入射光ベクトル】【正規化した光源方向を示すベクトル】
            NTLVector3 C = Pix.Camera; //【正規化した】
            NTLVector3 a, b, N, r;
            NTLRGB C1, C2, C3;
            int X,Y,j;
            float abs;
            float K1 = 0.7f;
            float K2 = 0.65f;
            float K3 = 0.65f;
            NTLRGB Cf = Pix.FaceColor;
            NTLRGB Cd = Pix.DiffuseColor;
            NTLRGB Cs = Pix.SpecularColor;
            float Pow = Pix.Power;
            float W;

            for(Y=1; Y<Dy-2; Y++)
            {
                for(X=1; X<Dx-2;X++)
                {
                    //【x方向, y方向の接線ベクトルa,bを求める】
                    a.X = Vertex[X + 1 + Dx * Y].X - Vertex[X - 1 + Dx * Y].X;
                    a.Y = 0.0f;
                    a.Z = Vertex[X + 1 + Dx * Y].Z - Vertex[X - 1 + Dx * Y].Z;     //距離を考慮すべき
                    a = VectorNormailze(a);
                    b.X = 0.0f;
                    b.Y = Vertex[X + Dx * (Y + 1)].Y - Vertex[X + Dx * (Y - 1)].Y;
                    b.Z = Vertex[X + Dx * (Y + 1)].Z - Vertex[X + Dx * (Y - 1)].Z; //距離を考慮すべき
                    b = VectorNormailze(b);
                    //【法線ベクトル】a, bに直交するベクトルを求める
                    N.X = a.Y * b.Z - a.Z * b.Y; N.Y = a.Z * b.X - a.X * b.Z; N.Z = a.X * b.Y - a.Y * b.X;
                    abs = 1.0f / (float)Math.Sqrt(N.X * N.X + N.Y * N.Y + N.Z * N.Z);
                    N.X *= abs; N.Y *= abs; N.Z *= abs; //*****************
                    //【物体色成分】一定
                    C1.R = K1 * Cf.R; C1.G = K1 * Cf.G; C1.B = K1 * Cf.B;
                    //【拡散反射成分】物体色を法線ベクトルNと入射光ベクトルLの内積値で決定
                    W = N.X * L.X + N.Y * L.Y + N.Z * L.Z;
                    if (W < 0.0f) { W = 0.0f; }
                    W *= K2; C2.R = W * Cd.R; C2.G = W * Cd.G; C2.B = W * Cd.B;
                    //【鏡面反射成分】
                    W = 2 * (L.X * N.X + L.Y * N.Y + L.Z * N.Z); //【反射光ベクトル】
                    r.X = W * N.X - L.X; r.Y = W * N.Y - L.Y; r.Z = W * N.Z - L.Z;
                    W = -(r.X * C.X + r.Y * C.Y + r.Z * C.Z);
                    if(W<0.0f) { W=0.0f; }
                    W = K3 * (float)Math.Pow(W, Pow);
                    C3.R = W * Cs.R; C3.G = W * Cs.G; C3.B = W * Cs.B;
                    //【効果】
                    j = (X + Dx * Y) * 4;
                    W = C1.B + C2.B + C3.B;
                    if( W > 1.0f) { W = 1.0f; }
                    dXRGB3[j] = (byte)(W * 255.0f);
                    W = C1.G + C2.G + C3.G;
                    if( W > 1.0f) { W = 1.0f; }
                    dXRGB3[j+1] = (byte)(W * 255.0f);
                    W = C1.R + C2.R + C3.R;
                    if (W > 1.0f) { W = 1.0f; }
                    dXRGB3[j+2] = (byte)(W * 255.0f);
                    dXRGB3[j+3] = 0;
                }
            }
            dBmpData = dBmp3.LockBits(new Rectangle(0, 0, Dx, Dy), ImageLockMode.WriteOnly, PixelFormat.Format32bppRgb);
            Marshal.Copy(dXRGB3, 0, dBmpData.Scan0, dXRGB3.Length); //【データ転送】メモリ⇒Bitmap
            dBmp3.UnlockBits(dBmpData);      //【Lock解除】
            dBmp3.RotateFlip(RotateFlipType.Rotate180FlipY); //【左右反転】
            pictureBox4.Image = dBmp3;
        }

        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        //【VectorNormailze】ベクトルの正規化
        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        NTLVector3 VectorNormailze(NTLVector3 v)
        {
            float a;
            a = 1.0f / (float)Math.Sqrt(v.X * v.X + v.Y * v.Y + v.Z * v.Z);
            v.X *= a; v.Y *= a; v.Z *= a;
            return v;
        }

        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        //【GetVectorLength】Vectorの長さ
        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        private float GetVectorLength(NTLVector3 a, NTLVector3 b)
        {
            float x, y, z, L;
            x = a.X - b.X; y = a.Y - b.Y; z = a.Z - b.Z;
            L = (float)Math.Sqrt(x * x + y * y + z * z);
            return L;
        }

        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        //【GetNormalVector】指定三角形の法線ベクトルを演算
        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        NTLVector3 GetNormalVector(NTLVector3 a, NTLVector3 b, NTLVector3 c)
        {
            NTLVector3 d, u, v; float abs;
            u.X = b.X - a.X; u.Y = b.Y - a.Y; u.Z = b.Z - a.Z; // u=b-a
            v.X = c.X - b.X; v.Y = c.Y - b.Y; v.Z = c.Z - b.Z; // v=c-b
            d.X = u.Y * v.Z - u.Z * v.Y;
            d.Y = u.Z * v.X - u.X * v.Z;
            d.Z = u.X * v.Y - u.Y * v.X;
            abs = 1.0f / (float)Math.Sqrt(d.X * d.X + d.Y * d.Y + d.Z * d.Z);
            d.X *= abs;
            d.Y *= abs;
            d.Z *= abs;
            return d;
        }

        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        //【GetVectorLength】Vectorの長さ
        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        void CreateSTLFile()
        {
            //【法線ベクトル】
            Xmin = Ymin = Zmin = +100000.0f;
            Xmax = Ymax = Zmax = -100000.0f;
            string FileName = "../../DepthImage.stl";
            Encoding utf8 = Encoding.GetEncoding(65001);

            CreateBaseIndices(); //まずポリゴンインデックスを作成

            System.IO.File.Delete(FileName);
            StreamWriter writer = new StreamWriter(FileName, true, utf8);
            writer.WriteLine("solid NTL_KinectFusion");
            int p1, p2, p3; NTLVector3 a, b, c, d;
            for (int i = 0; i < nIndex; i += 3)
            {
                p1 = Indices[i]; p2 = Indices[i + 1]; p3 = Indices[i + 2];
                a = Vertex[p1]; b = Vertex[p2]; c = Vertex[p3];
                if (Xmax < a.X) { Xmax = a.X; }
                if (Xmax < b.X) { Xmax = b.X; }
                if (Xmax < c.X) { Xmax = c.X; }
                if (Xmin > a.X) { Xmin = a.X; }
                if (Xmin > b.X) { Xmin = b.X; }
                if (Xmin > c.X) { Xmin = c.X; }
                if (Ymax < a.Y) { Ymax = a.Y; }
                if (Ymax < b.Y) { Ymax = b.Y; }
                if (Ymax < c.Y) { Ymax = c.Y; }
                if (Ymin > a.Y) { Ymin = a.Y; }
                if (Ymin > b.Y) { Ymin = b.Y; }
                if (Ymin > c.Y) { Ymin = c.Y; }
                if (Zmax < a.Z) { Zmax = a.Z; }
                if (Zmax < b.Z) { Zmax = b.Z; }
                if (Zmax < c.Z) { Zmax = c.Z; }
                if (Zmin > a.Z) { Zmin = a.Z; }
                if (Zmin > b.Z) { Zmin = b.Z; }
                if (Zmin > c.Z) { Zmin = c.Z; }
                d = GetNormalVector(a, b, c);
                writer.WriteLine("facet normal " + d.X.ToString() + " " + d.Y.ToString() + " " + d.Y.ToString() + " ");
                writer.WriteLine("   outer loop ");
                writer.WriteLine("     vertex " + a.X.ToString() + " " + a.Y.ToString() + " " + a.Z.ToString() + " ");
                writer.WriteLine("     vertex " + b.X.ToString() + " " + b.Y.ToString() + " " + b.Z.ToString() + " ");
                writer.WriteLine("     vertex " + c.X.ToString() + " " + c.Y.ToString() + " " + c.Z.ToString() + " ");
                writer.WriteLine("   endloop ");
                writer.WriteLine(" endface ");
            }
            writer.WriteLine("endsolid NTL_KinectFusion");
            writer.Close();
        }

        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        //【CreateIndices】
        //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
        void CreateBaseIndices()
        {
            int p1, p2, p3, p4, X, Y;
            float L1, L2, L3;
            NTLVector3 a, b, c, d;

            //【ポリゴンデータ作成】≪インデックスデータ≫
            nIndex = 0; nPolygon = 0;
            Rectangle r = new Rectangle(1, 1, 510, 422);
            for (Y = r.Y; Y < (r.Y + r.Height - 1); Y++)
            {
                for (X = r.X; X < (r.X + r.Width - 1); X++)
                {
                    p1 = X + Dx * Y; p2 = X + Dx * (Y + 1); p3 = (X + 1) + Dx * Y; p4 = (X + 1) + Dx * (Y + 1);
                    a = Vertex[p1]; b = Vertex[p2]; c = Vertex[p3]; d = Vertex[p4];
                    L1 = GetVectorLength(a, b); L2 = GetVectorLength(b, c); L3 = GetVectorLength(c, a);
                    if ((Vertex[p1].Z > 0.5f) && (Vertex[p2].Z > 0.5f) && (Vertex[p3].Z > 0.5f) &&
                        (L1 < 0.05f) && (L2 < 0.05f) && (L3 < 0.05f)) //【一辺の長さが5cm以上なら生成しない】
                    {
                        Indices[nIndex] = p1; Indices[nIndex + 1] = p2; Indices[nIndex + 2] = p3;
                        nIndex += 3; nPolygon++;
                    }
                    L1 = GetVectorLength(d, c); L2 = GetVectorLength(c, b); L3 = GetVectorLength(b, d);
                    if ((Vertex[p4].Z > 0.5f) && (Vertex[p3].Z > 0.5f) && (Vertex[p2].Z > 0.5f) &&
                        (L1 < 0.05f) && (L2 < 0.05f) && (L3 < 0.05f)) //【一辺の長さが5cm以上なら生成しない】
                    {
                        Indices[nIndex] = p4; Indices[nIndex + 1] = p3; Indices[nIndex + 2] = p2;
                        nIndex += 3; nPolygon++;
                    }
                }
            }
        }

    }
}
